import numpy

from thinc.backends._param_server import ParamServer


def test_param_server_init():
    array = numpy.zeros((5,), dtype="f")
    params = {("a", 1): array, ("b", 2): array}
    grads = {("a", 1): array, ("c", 3): array}
    ps = ParamServer(params, grads)
    assert ps.param_keys == (("a", 1), ("b", 2))
    assert ps.grad_keys == (("a", 1),)



from functools import partial

import pytest

from thinc.api import Linear, resizable
from thinc.layers.resizable import resize_linear_weighted, resize_model


@pytest.fixture
def model():
    output_layer = Linear(nO=None, nI=None)
    fill_defaults = {"b": 0, "W": 0}
    model = resizable(
        output_layer,
        resize_layer=partial(resize_linear_weighted, fill_defaults=fill_defaults),
    )
    return model


def test_resizable_linear_default_name(model):
    assert model.name == "resizable(linear)"


def test_resize_model(model):
    """Test that resizing the model doesn't cause an exception."""
    resize_model(model, new_nO=10)
    resize_model(model, new_nO=11)

    model.set_dim("nO", 0, force=True)
    resize_model(model, new_nO=10)

    model.set_dim("nI", 10, force=True)
    model.set_dim("nO", 0, force=True)
    resize_model(model, new_nO=10)



import numpy
import pytest
from hypothesis import given

from thinc.api import Padded, Ragged, get_width
from thinc.types import ArgsKwargs
from thinc.util import (
    convert_recursive,
    get_array_module,
    is_cupy_array,
    is_numpy_array,
    to_categorical,
)

from . import strategies

ALL_XP = [numpy]
try:
    import cupy

    ALL_XP.append(cupy)
except ImportError:
    pass


@pytest.mark.parametrize(
    "obj,width",
    [
        (numpy.zeros((1, 2, 3, 4)), 4),
        (numpy.array(1), 0),
        (numpy.array([1, 2]), 3),
        ([numpy.zeros((1, 2)), numpy.zeros((1))], 2),
        (Ragged(numpy.zeros((1, 2)), numpy.zeros(1)), 2),  # type:ignore
        (
            Padded(
                numpy.zeros((2, 1, 2)),  # type:ignore
                numpy.zeros(2),  # type:ignore
                numpy.array([1, 0]),  # type:ignore
                numpy.array([0, 1]),  # type:ignore
            ),
            2,
        ),
        ([], 0),
    ],
)
def test_get_width(obj, width):
    assert get_width(obj) == width


@pytest.mark.parametrize("obj", [1234, "foo", {"a": numpy.array(0)}])
def test_get_width_fail(obj):
    with pytest.raises(ValueError):
        get_width(obj)


@pytest.mark.parametrize("xp", ALL_XP)
def test_array_module_cpu_gpu_helpers(xp):
    error = (
        "Only numpy and cupy arrays are supported"
        ", but found <class 'int'> instead. If "
        "get_array_module module wasn't called "
        "directly, this might indicate a bug in Thinc."
    )
    with pytest.raises(ValueError, match=error):
        get_array_module(0)
    zeros = xp.zeros((1, 2))
    xp_ = get_array_module(zeros)
    assert xp_ == xp
    if xp == numpy:
        assert is_numpy_array(zeros)
        assert not is_numpy_array((1, 2))
    else:
        assert is_cupy_array(zeros)
        assert not is_cupy_array((1, 2))


@given(
    label_smoothing=strategies.floats(min_value=0.0, max_value=0.5, exclude_max=True)
)
def test_to_categorical(label_smoothing):
    # Test without n_classes
    one_hot = to_categorical(numpy.asarray([1, 2], dtype="i"))
    assert one_hot.shape == (2, 3)
    # From keras
    # https://github.com/keras-team/keras/blob/master/tests/keras/utils/np_utils_test.py
    nc = 5
    shapes = [(1,), (3,), (4, 3), (5, 4, 3), (3, 1), (3, 2, 1)]
    expected_shapes = [
        (1, nc),
        (3, nc),
        (4, 3, nc),
        (5, 4, 3, nc),
        (3, 1, nc),
        (3, 2, 1, nc),
    ]
    labels = [numpy.random.randint(0, nc, shape) for shape in shapes]
    one_hots = [to_categorical(label, nc) for label in labels]
    smooths = [
        to_categorical(label, nc, label_smoothing=label_smoothing) for label in labels
    ]
    for i in range(len(expected_shapes)):
        label = labels[i]
        one_hot = one_hots[i]
        expected_shape = expected_shapes[i]
        smooth = smooths[i]
        assert one_hot.shape == expected_shape
        assert smooth.shape == expected_shape
        assert numpy.array_equal(one_hot, one_hot.astype(bool))
        assert numpy.all(one_hot.sum(axis=-1) == 1)
        assert numpy.all(numpy.argmax(one_hot, -1).reshape(label.shape) == label)
        assert numpy.all(smooth.argmax(axis=-1) == one_hot.argmax(axis=-1))
        assert numpy.all(numpy.isclose(numpy.sum(smooth, axis=-1), 1.0))
        assert numpy.isclose(numpy.max(smooth), 1 - label_smoothing)
        assert numpy.isclose(
            numpy.min(smooth), label_smoothing / (smooth.shape[-1] - 1)
        )

    # At least one class is required without label smoothing.
    numpy.testing.assert_allclose(
        to_categorical(numpy.asarray([0, 0, 0]), 1), [[1.0], [1.0], [1.0]]
    )
    numpy.testing.assert_allclose(
        to_categorical(numpy.asarray([0, 0, 0])), [[1.0], [1.0], [1.0]]
    )
    with pytest.raises(ValueError, match=r"n_classes should be at least 1"):
        to_categorical(numpy.asarray([0, 0, 0]), 0)

    # At least two classes are required with label smoothing.
    numpy.testing.assert_allclose(
        to_categorical(numpy.asarray([0, 1, 0]), 2, label_smoothing=0.01),
        [[0.99, 0.01], [0.01, 0.99], [0.99, 0.01]],
    )
    numpy.testing.assert_allclose(
        to_categorical(numpy.asarray([0, 1, 0]), label_smoothing=0.01),
        [[0.99, 0.01], [0.01, 0.99], [0.99, 0.01]],
    )
    with pytest.raises(
        ValueError, match=r"n_classes should be greater than 1.*label smoothing.*but 1"
    ):
        to_categorical(numpy.asarray([0, 1, 0]), 1, label_smoothing=0.01),
    with pytest.raises(
        ValueError, match=r"n_classes should be greater than 1.*label smoothing.*but 1"
    ):
        to_categorical(numpy.asarray([0, 0, 0]), label_smoothing=0.01),

    with pytest.raises(ValueError, match=r"label_smoothing parameter"):
        to_categorical(numpy.asarray([0, 1, 2, 3, 4]), label_smoothing=0.8)

    with pytest.raises(ValueError, match=r"label_smoothing parameter"):
        to_categorical(numpy.asarray([0, 1, 2, 3, 4]), label_smoothing=0.88)


def test_convert_recursive():
    is_match = lambda obj: obj == "foo"
    convert_item = lambda obj: obj.upper()
    obj = {
        "a": {("b", "foo"): {"c": "foo", "d": ["foo", {"e": "foo", "f": (1, "foo")}]}}
    }
    result = convert_recursive(is_match, convert_item, obj)
    assert result["a"][("b", "FOO")]["c"] == "FOO"
    assert result["a"][("b", "FOO")]["d"] == ["FOO", {"e": "FOO", "f": (1, "FOO")}]
    obj = {"a": ArgsKwargs(("foo", [{"b": "foo"}]), {"a": ["x", "foo"]})}
    result = convert_recursive(is_match, convert_item, obj)
    assert result["a"].args == ("FOO", [{"b": "FOO"}])
    assert result["a"].kwargs == {"a": ["x", "FOO"]}



import ast
import importlib
from collections import namedtuple
from typing import List, Tuple

import pytest

_Import = namedtuple("_Import", ["module", "name", "alias"])


def get_imports(path: str) -> Tuple[_Import, ...]:
    """Parse Python file at path, retrieve import statements.
    Adapted from https://stackoverflow.com/a/9049549.
    path (str): Path to Python file.
    RETURNS (Tuple[_Import]): All imports found in file at path.
    """
    with open(path) as fh:
        root = ast.parse(fh.read(), path)

    imports: List[_Import] = []
    for node in ast.walk(root):
        if isinstance(node, ast.Import):
            module: List[str] = []
        elif isinstance(node, ast.ImportFrom) and node.module:
            module = node.module.split(".")
        else:
            continue

        assert isinstance(node, (ast.Import, ast.ImportFrom))
        imports.extend(
            [_Import(module, n.name.split("."), n.asname) for n in node.names]
        )

    return tuple(imports)


@pytest.mark.parametrize("module_name", ["thinc.api", "thinc.shims", "thinc.layers"])
def test_import_reexport_equivalency(module_name: str):
    """Tests whether a module's __all__ is equivalent to its imports. This assumes that this module is supposed to
    re-export all imported values.
    module_name (str): Module to load.
    """
    mod = importlib.import_module(module_name)

    assert set(mod.__all__) == {
        k
        for k in set(n for i in get_imports(str(mod.__file__)) for n in i.name)
        if (
            # Ignore all values prefixed with _, as we expect those not to be re-exported.
            # However, __version__ should be reexported in thinc/__init__.py.
            (not k.startswith("_") or module_name == "thinc" and k == "__version__")
        )
    }



from typing import Any, Callable, List, Optional, Sequence, Tuple, TypeVar, cast

from ..config import registry
from ..model import Model

InItemT = TypeVar("InItemT")
OutItemT = TypeVar("OutItemT")
ItemT = TypeVar("ItemT")

NestedT = List[List[ItemT]]
FlatT = List[ItemT]


@registry.layers("with_flatten.v2")
def with_flatten_v2(
    layer: Model[FlatT[InItemT], FlatT[OutItemT]]
) -> Model[NestedT[InItemT], NestedT[OutItemT]]:
    return Model(f"with_flatten({layer.name})", forward, layers=[layer], init=init)


def forward(
    model: Model[NestedT[InItemT], NestedT[OutItemT]],
    Xnest: NestedT[InItemT],
    is_train: bool,
) -> Tuple[NestedT[OutItemT], Callable]:
    layer: Model[FlatT[InItemT], FlatT[OutItemT]] = model.layers[0]
    Xflat, lens = _flatten(Xnest)
    Yflat, backprop_layer = layer(Xflat, is_train)
    Ynest = _unflatten(Yflat, lens)

    def backprop(dYnest: NestedT[InItemT]) -> NestedT[OutItemT]:
        dYflat, _ = _flatten(dYnest)  # type: ignore[arg-type, var-annotated]
        # type ignore necessary for older versions of Mypy/Pydantic
        dXflat = backprop_layer(dYflat)
        dXnest = _unflatten(dXflat, lens)
        return dXnest

    return Ynest, backprop


def _flatten(nested: NestedT[ItemT]) -> Tuple[FlatT[ItemT], List[int]]:
    flat: List = []
    lens: List[int] = []
    for item in nested:
        flat.extend(item)
        lens.append(len(item))
    return cast(FlatT[ItemT], flat), lens


def _unflatten(flat: FlatT[ItemT], lens: List[int]) -> NestedT[ItemT]:
    nested = []
    for l in lens:
        nested.append(flat[:l])
        flat = flat[l:]
    return nested


def init(
    model: Model[NestedT[InItemT], NestedT[OutItemT]],
    X: Optional[NestedT[InItemT]] = None,
    Y: Optional[NestedT[OutItemT]] = None,
) -> None:
    model.layers[0].initialize(
        _flatten(X)[0] if X is not None else None,
        model.layers[0].ops.xp.hstack(Y) if Y is not None else None,
    )



import pytest

from thinc.api import (
    ParametricAttention,
    Relu,
    Softmax,
    chain,
    list2ragged,
    reduce_max,
    reduce_sum,
    with_ragged,
)
from thinc.util import DataValidationError, data_validation


@pytest.mark.xfail(reason="Validation currently disabled for Pydantic 2 changes0")
def test_validation():
    model = chain(Relu(10), Relu(10), with_ragged(reduce_max()), Softmax())
    with data_validation(True):
        with pytest.raises(DataValidationError):
            model.initialize(X=model.ops.alloc2f(1, 10), Y=model.ops.alloc2f(1, 10))
        with pytest.raises(DataValidationError):
            model.initialize(X=model.ops.alloc3f(1, 10, 1), Y=model.ops.alloc2f(1, 10))
        with pytest.raises(DataValidationError):
            model.initialize(X=[model.ops.alloc2f(1, 10)], Y=model.ops.alloc2f(1, 10))


@pytest.mark.xfail(reason="Validation currently disabled for Pydantic 2 changes0")
def test_validation_complex():
    good_model = chain(list2ragged(), reduce_sum(), Relu(12, dropout=0.5), Relu(1))
    X = [good_model.ops.xp.zeros((4, 75), dtype="f")]
    Y = good_model.ops.xp.zeros((1,), dtype="f")
    good_model.initialize(X, Y)
    good_model.predict(X)

    bad_model = chain(
        list2ragged(),
        reduce_sum(),
        Relu(12, dropout=0.5),
        # ERROR: Why can't I attach a Relu to an attention layer?
        ParametricAttention(12),
        Relu(1),
    )
    with data_validation(True):
        with pytest.raises(DataValidationError):
            bad_model.initialize(X, Y)



from typing import Callable, Optional, Tuple

import numpy

from ..config import registry
from ..model import Model
from ..types import Floats2d, Ints2d

InT = Ints2d
OutT = Floats2d


@registry.layers("uniqued.v1")
def uniqued(layer: Model, *, column: int = 0) -> Model[InT, OutT]:
    """Group inputs to a layer, so that the layer only has to compute for the
    unique values. The data is transformed back before output, and the same
    transformation is applied for the gradient. Effectively, this is a cache
    local to each minibatch.
    """
    return Model(
        f"uniqued({layer.name})",
        forward,
        init=init,
        layers=[layer],
        dims={"nO": None, "nI": None},
        attrs={"column": column},
    )


def forward(model: Model[InT, OutT], X: InT, is_train: bool) -> Tuple[OutT, Callable]:
    column: int = model.attrs["column"]
    layer = model.layers[0]
    if X.size < 2:
        return layer(X, is_train)
    keys = X[:, column]
    if not isinstance(keys, numpy.ndarray):
        keys = keys.get()  # pragma: no cover
    uniq_keys, ind, inv, counts = layer.ops.xp.unique(
        keys, return_index=True, return_inverse=True, return_counts=True
    )
    counts = model.ops.reshape2i(counts, -1, 1)
    X_uniq = X[ind]
    Y_uniq, bp_Y_uniq = layer(X_uniq, is_train)
    Y = Y_uniq[inv].reshape((X.shape[0],) + Y_uniq.shape[1:])
    uniq_shape = tuple(Y_uniq.shape)

    def backprop(dY: OutT) -> InT:
        dY_uniq = layer.ops.alloc2f(*uniq_shape)
        layer.ops.scatter_add(dY_uniq, layer.ops.asarray_i(inv), dY)
        d_uniques = bp_Y_uniq(dY_uniq)
        # This confusing bit of indexing "ununiques"
        return (d_uniques / counts)[inv]

    return Y, backprop


def init(
    model: Model[InT, OutT], X: Optional[InT] = None, Y: Optional[OutT] = None
) -> None:
    layer = model.layers[0]
    layer.initialize(X=X, Y=Y)
    if layer.has_dim("nI"):
        model.set_dim("nI", layer.get_dim("nI"))  # pragma: no cover
    if layer.has_dim("nO"):
        model.set_dim("nO", layer.get_dim("nO"))



"""Generators that provide different rates, schedules, decays or series."""
from typing import Iterable

import numpy

from .config import registry


@registry.schedules("constant_then.v1")
def constant_then(
    rate: float, steps: int, schedule: Iterable[float]
) -> Iterable[float]:
    """Yield a constant rate for N steps, before starting a schedule."""
    for i in range(steps):
        yield rate
    for value in schedule:
        yield value


@registry.schedules("constant.v1")
def constant(rate: float) -> Iterable[float]:
    """Yield a constant rate."""
    while True:
        yield rate


@registry.schedules("decaying.v1")
def decaying(base_rate: float, decay: float, *, t: int = 0) -> Iterable[float]:
    """Yield an infinite series of linearly decaying values,
    following the schedule: base_rate * 1 / (1 + decay * t)

    EXAMPLE:
        >>> learn_rates = decaying(0.001, 1e-4)
        >>> next(learn_rates)
        0.001
        >>> next(learn_rates)
        0.00999
    """
    while True:
        yield base_rate * (1.0 / (1.0 + decay * t))
        t += 1


@registry.schedules("compounding.v1")
def compounding(
    start: float, stop: float, compound: float, *, t: float = 0.0
) -> Iterable[float]:
    """Yield an infinite series of compounding values. Each time the
    generator is called, a value is produced by multiplying the previous
    value by the compound rate.

    EXAMPLE:
        >>> sizes = compounding(1.0, 10.0, 1.5)
        >>> assert next(sizes) == 1.
        >>> assert next(sizes) == 1 * 1.5
        >>> assert next(sizes) == 1.5 * 1.5
    """
    curr = float(start)
    while True:
        yield _clip(curr, start, stop)
        curr *= compound


def _clip(value: float, start: float, stop: float) -> float:
    return max(value, stop) if (start > stop) else min(value, stop)


@registry.schedules("slanted_triangular.v1")
def slanted_triangular(
    max_rate: float,
    num_steps: int,
    *,
    cut_frac: float = 0.1,
    ratio: int = 32,
    decay: float = 1.0,
    t: float = 0.0,
) -> Iterable[float]:
    """Yield an infinite series of values according to Howard and Ruder's
    "slanted triangular learning rate" schedule.
    """
    cut = int(num_steps * cut_frac)
    while True:
        t += 1
        if t < cut:
            p = t / cut
        else:
            p = 1 - ((t - cut) / (cut * (1 / cut_frac - 1)))
        learn_rate = max_rate * (1 + p * (ratio - 1)) * (1 / ratio)
        yield learn_rate


@registry.schedules("warmup_linear.v1")
def warmup_linear(
    initial_rate: float, warmup_steps: int, total_steps: int
) -> Iterable[float]:
    """Generate a series, starting from an initial rate, and then with a warmup
    period, and then a linear decline. Used for learning rates.
    """
    step = 0
    while True:
        if step < warmup_steps:
            factor = step / max(1, warmup_steps)
        else:
            factor = max(
                0.0, (total_steps - step) / max(1.0, total_steps - warmup_steps)
            )
        yield factor * initial_rate
        step += 1


@registry.schedules("cyclic_triangular.v1")
def cyclic_triangular(min_lr: float, max_lr: float, period: int) -> Iterable[float]:
    it = 1
    while True:
        # https://towardsdatascience.com/adaptive-and-cyclical-learning-rates-using-pytorch-2bf904d18dee
        cycle = numpy.floor(1 + it / (2 * period))
        x = numpy.abs(it / period - 2 * cycle + 1)
        relative = max(0, 1 - x)
        yield min_lr + (max_lr - min_lr) * relative
        it += 1


__all__ = [
    "cyclic_triangular",
    "warmup_linear",
    "constant",
    "constant_then",
    "decaying",
    "warmup_linear",
    "slanted_triangular",
    "compounding",
]



from typing import Any, Callable, Optional, Tuple, TypeVar

from ..model import Model

_ModelT = TypeVar("_ModelT", bound=Model)

do_nothing = lambda *args, **kwargs: None


def with_debug(
    layer: _ModelT,
    name: Optional[str] = None,
    *,
    on_init: Callable[[Model, Any, Any], None] = do_nothing,
    on_forward: Callable[[Model, Any, bool], None] = do_nothing,
    on_backprop: Callable[[Any], None] = do_nothing,
) -> _ModelT:
    """Debugging layer that wraps any layer and allows executing callbacks
    during the forward pass, backward pass and initialization. The callbacks
    will receive the same arguments as the functions they're called in.
    """
    name = layer.name if name is None else name

    orig_forward = layer._func
    orig_init = layer.init

    def forward(model: Model, X: Any, is_train: bool) -> Tuple[Any, Callable]:
        on_forward(model, X, is_train)
        layer_Y, layer_callback = orig_forward(layer, X, is_train=is_train)

        def backprop(dY: Any) -> Any:
            on_backprop(dY)
            return layer_callback(dY)

        return layer_Y, backprop

    def init(model: Model, X: Any, Y: Any) -> None:
        on_init(model, X, Y)
        if orig_init is not None:
            orig_init(layer, X, Y)

    layer.replace_callbacks(forward, init=init)

    return layer



import numpy
import pytest

from thinc.api import (
    PyTorchWrapper_v2,
    TorchScriptWrapper_v1,
    pytorch_to_torchscript_wrapper,
)
from thinc.compat import has_torch, torch


@pytest.mark.skipif(not has_torch, reason="needs PyTorch")
@pytest.mark.parametrize("nN,nI,nO", [(2, 3, 4)])
def test_pytorch_script(nN, nI, nO):

    model = PyTorchWrapper_v2(torch.nn.Linear(nI, nO)).initialize()
    script_model = pytorch_to_torchscript_wrapper(model)

    X = numpy.random.randn(nN, nI).astype("f")
    Y = model.predict(X)
    Y_script = script_model.predict(X)
    numpy.testing.assert_allclose(Y, Y_script)

    serialized = script_model.to_bytes()
    script_model2 = TorchScriptWrapper_v1()
    script_model2.from_bytes(serialized)

    numpy.testing.assert_allclose(Y, script_model2.predict(X))



import catalogue
import confection
from confection import VARIABLE_RE, Config, ConfigValidationError, Promise

from .types import Decorator


class registry(confection.registry):
    # fmt: off
    optimizers: Decorator = catalogue.create("thinc", "optimizers", entry_points=True)
    schedules: Decorator = catalogue.create("thinc", "schedules", entry_points=True)
    layers: Decorator = catalogue.create("thinc", "layers", entry_points=True)
    losses: Decorator = catalogue.create("thinc", "losses", entry_points=True)
    initializers: Decorator = catalogue.create("thinc", "initializers", entry_points=True)
    datasets: Decorator = catalogue.create("thinc", "datasets", entry_points=True)
    ops: Decorator = catalogue.create("thinc", "ops", entry_points=True)
    # fmt: on

    @classmethod
    def create(cls, registry_name: str, entry_points: bool = False) -> None:
        """Create a new custom registry."""
        if hasattr(cls, registry_name):
            raise ValueError(f"Registry '{registry_name}' already exists")
        reg: Decorator = catalogue.create(
            "thinc", registry_name, entry_points=entry_points
        )
        setattr(cls, registry_name, reg)


__all__ = ["Config", "registry", "ConfigValidationError", "Promise", "VARIABLE_RE"]



from thinc.api import Linear, chain


def test_issue208():
    """Test issue that was caused by trying to flatten nested chains."""
    layer1 = Linear(nO=9, nI=3)
    layer2 = Linear(nO=12, nI=9)
    layer3 = Linear(nO=5, nI=12)
    model = chain(layer1, chain(layer2, layer3)).initialize()
    assert model.get_dim("nO") == 5



from typing import Callable, Tuple, TypeVar, cast

from ..config import registry
from ..model import Model
from ..types import FloatsXdT


@registry.layers("sigmoid_activation.v1")
def sigmoid_activation() -> Model[FloatsXdT, FloatsXdT]:
    return Model("sigmoid_activation", forward)


def forward(
    model: Model[FloatsXdT, FloatsXdT], X: FloatsXdT, is_train: bool
) -> Tuple[FloatsXdT, Callable]:
    Y = model.ops.sigmoid(X, inplace=False)

    def backprop(dY: FloatsXdT) -> FloatsXdT:
        return cast(
            FloatsXdT,
            dY * model.ops.dsigmoid(Y, inplace=False),  # type:ignore[operator]
        )

    return Y, backprop



import numpy

from thinc.api import HashEmbed


def test_init():
    model = HashEmbed(64, 1000).initialize()
    assert model.get_dim("nV") == 1000
    assert model.get_dim("nO") == 64
    assert model.get_param("E").shape == (1000, 64)


def test_seed_changes_bucket():
    model1 = HashEmbed(64, 1000, seed=2).initialize()
    model2 = HashEmbed(64, 1000, seed=1).initialize()
    arr = numpy.ones((1,), dtype="uint64")
    vector1 = model1.predict(arr)
    vector2 = model2.predict(arr)
    assert vector1.sum() != vector2.sum()



import pytest
from hypothesis import given, settings
from hypothesis.strategies import lists, one_of, tuples

from thinc.api import PyTorchGradScaler
from thinc.compat import has_torch, has_torch_amp, has_torch_cuda_gpu, torch
from thinc.util import is_torch_array

from ..strategies import ndarrays


def tensors():
    return ndarrays().map(lambda a: torch.tensor(a).cuda())


@pytest.mark.skipif(not has_torch, reason="needs PyTorch")
@pytest.mark.skipif(not has_torch_cuda_gpu, reason="needs a GPU")
@pytest.mark.skipif(
    not has_torch_amp, reason="requires PyTorch with mixed-precision support"
)
@given(X=one_of(tensors(), lists(tensors()), tuples(tensors())))
@settings(deadline=None)
def test_scale_random_inputs(X):
    import torch

    device_id = torch.cuda.current_device()
    scaler = PyTorchGradScaler(enabled=True)
    scaler.to_(device_id)

    if is_torch_array(X):
        assert torch.allclose(scaler.scale(X), X * 2.0**16)
    else:
        scaled1 = scaler.scale(X)
        scaled2 = [t * 2.0**16 for t in X]
        for t1, t2 in zip(scaled1, scaled2):
            assert torch.allclose(t1, t2)


@pytest.mark.skipif(not has_torch, reason="needs PyTorch")
@pytest.mark.skipif(not has_torch_cuda_gpu, reason="needs a GPU")
@pytest.mark.skipif(
    not has_torch_amp, reason="requires PyTorch with mixed-precision support"
)
def test_grad_scaler():
    import torch

    device_id = torch.cuda.current_device()

    scaler = PyTorchGradScaler(enabled=True)
    scaler.to_(device_id)

    #  Test that scaling works as expected.
    t = torch.tensor([1.0], device=device_id)
    assert scaler.scale([torch.tensor([1.0], device=device_id)]) == [
        torch.tensor([2.0**16], device=device_id)
    ]
    assert scaler.scale(torch.tensor([1.0], device=device_id)) == torch.tensor(
        [2.0**16], device=device_id
    )
    with pytest.raises(ValueError):
        scaler.scale("bogus")
    with pytest.raises(ValueError):
        scaler.scale(42)

    # Test infinity detection.
    g = [
        torch.tensor([2.0**16], device=device_id),
        torch.tensor([float("Inf")], device=device_id),
    ]

    # Check that infinity was found.
    assert scaler.unscale(g)

    # Check whether unscale was successful.
    assert g[0] == torch.tensor([1.0]).cuda()

    scaler.update()

    # Since infinity was found, the scale should be halved from 2**16
    # to 2**15 for the next step.
    assert scaler.scale([torch.tensor([1.0], device=device_id)]) == [
        torch.tensor([2.0**15], device=device_id)
    ]


@pytest.mark.skipif(not has_torch, reason="needs PyTorch")
@pytest.mark.skipif(
    has_torch_amp, reason="needs PyTorch without gradient scaling support"
)
def test_raises_on_old_pytorch():
    import torch

    scaler = PyTorchGradScaler(enabled=True)
    with pytest.raises(ValueError, match=r"not supported.*1.9.0"):
        scaler.scale([torch.tensor([1.0], device="cpu")])


@pytest.mark.skipif(not has_torch, reason="needs PyTorch")
@pytest.mark.skipif(
    not has_torch_amp, reason="needs PyTorch with gradient scaling support"
)
def test_raises_with_cpu_tensor():
    import torch

    scaler = PyTorchGradScaler(enabled=True)
    with pytest.raises(
        ValueError, match=r"Gradient scaling is only supported for CUDA tensors."
    ):
        scaler.scale([torch.tensor([1.0], device="cpu")])
